/*

	     Chi-square distribution calculation program

		 by John Walker (kelvin@fourmilab.ch)
			     January 1997

		This program is in the public domain.

    This program allows calculation of any of the components
    of the chi-square distribution from a given values for an
    adequate set of other components.  This is simply a command
    line front-end; all calculation is done by calling DCDFLIB.

    Components:

	P   Integral from 0 to X of the chi-square distribution.
	Q   1 - P.
	X   Upper limit of integration of the chi-square distribution.
	D   Degrees of freedom of the chi-square distribution.

    each known component is specified as the corresponding option
    on the command line, followed immediately by its value, which
    is a floating point number which can be in scientific notation
    acceptable to the C library function atof().
 
    To calculate Q:

	  chisquare -Xnnn -Dnnn

    To calculate P, calculate Q and subtract it from 1 (duh).

    To calculate X:

	  chisquare -Pnnn -Dnnn
      or: chisquare -Qnnn -Dnnn

    To calculate D:

	  chisquare -Pnnn -Xnnn
      or: chisquare -Qnnn -Xnnn

    For example, you've thrown a given die 1000 times, tabulated
    the results, and calculated a chi-square value of 20.5 based
    on the probability table for a fair die (each face has a
    probability of 1/6 of coming up).  To calculate the probability
    the die is in fact fair, use:

	  chisquare -D5 -X20.5

    which prints a Q value of 0.00100652778639 on standard output.
    The odds of the die being fair are, thus, about one in a thousand
    (or, phrased another way, you'd have to almost a thousand series
    of 1000 casts before you would expect another result with a
    chi-square value this high.  Note that the D value, "degrees
    of freedom", is one less than the number of possible outcomes
    of the experiment.	For coin flipping, for example, D is 1.

    The return value is 0 if a valid value was calculated, 2 if
    insufficient or over-constrained arguments were supplied, and
    -1 if an error occurred in the calculation (usually because
    out-of-range arguments were supplied).

    The -L option causes all arguments, input and output to be
    printed, separated by spaces, in the order:

	  D P Q X

    for example:

	  chisquare -D5 -X20.5 -L

    prints:

	  5 0.998993472214 0.00100652778639 20.5

*/


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <values.h>
#include "dcdflib.h"

/*  USAGE  --  Print how-to-call information.  */

static void usage()
{
    fprintf(stderr,"CHISQ  --  Calculate chi-square disribution.  Call");
    fprintf(stderr,
       "\n           with chisq options...");
    fprintf(stderr,"\n");
    fprintf(stderr,"\n     Options:");
    fprintf(stderr,"\n               -Dnum  Degrees of freedom");
    fprintf(stderr,"\n               -L     Output D P Q X on one line");
    fprintf(stderr,"\n               -Pnum  Integral from 0 to X of chi-square distribution");
    fprintf(stderr,"\n               -Q     1 - P");
    fprintf(stderr,"\n               -X     Upper limit of integration of chi-square distribution");
    fprintf(stderr,"\n               -U     Print this message");
    fprintf(stderr,"\n");
}

/*  Main program   */

int main(argc, argv)
  int argc; char *argv[];
{
    int i, all = 0;
    char *cp, opt;

#define UNSPEC	MINDOUBLE
#define SPEC(x) ((x) != UNSPEC)

    int which, status;
    double p = UNSPEC, q = UNSPEC, df = UNSPEC, x = UNSPEC, bound, result;

    for (i = 1; i < argc; i++) {
	cp = argv[i];
        if (*cp == '-') {
	    opt = *(++cp);
	    if (islower(opt)) {
		opt = toupper(opt);
	    }
	    switch (opt) {

		/*  Input function arguments  */

                case 'D':
		    df = atof(cp + 1);
		    break;

                case 'P':
		    p = atof(cp + 1);
		    break;

                case 'Q':
		    q = atof(cp + 1);
		    break;

                case 'X':
		    x = atof(cp + 1);
		    break;

		/*  Output format options  */

                case 'L':
		    all = 1;
		    break;

                case '?':
                case 'U':
		    usage();
		    return 0;
	    }
	}
    }

    if (SPEC(p) && SPEC(q) && (q != (1 - p))) {
        printf("Over-constrained: P and Q both specified, Q != (1 - P).\n");
	return 2;
    }

    if (SPEC(x) && SPEC(df) && (SPEC(p) || SPEC(q))) {
        printf("Over-constrained: X and DF specified with P or Q.\n");
	return 2;
    }

    if (SPEC(x) && SPEC(df) && (SPEC(p) || SPEC(q))) {
        printf("Over-constrained: X and DF specified with P or Q.\n");
	return 2;
    }

    if (SPEC(p) && !SPEC(q)) {
	q = 1 - p;
    } else if (SPEC(q) && !SPEC(p)) {
	p = 1 - q;
    }

    if (SPEC(x) && SPEC(df)) {
	which = 1;
    } else if (SPEC(p) && SPEC(df)) {
	which = 2;
    } else if (SPEC(p) && SPEC(x)) {
	which = 3;
    } else {
	if (SPEC(x) || SPEC(df) || SPEC(p) || SPEC(x)) {
            printf("Under-constrained.  Must specify one of:\n");
            printf("  -X and -DF to calculate Q\n");
            printf("  -DF and -P or -Q to calculate X\n");
            printf("  -X and -P or -Q to calculate DF\n\n");
	}
	usage();
	return 2;
    }

    cdfchi(&which, &p, &q, &x, &df, &status, &bound);

    if (status == 0) {
	if (all) {
            printf("%.12g %.12g %.12g %.12g\n", df, p, q, x);
	} else {
	    switch (which) {
		case 1:
		    result = q;
		    break;

		case 2:
		    result = x;
		    break;

		case 3:
		    result = df;
		    break;
	    }
            printf("%.12g\n", result);
	}
    } else {
	switch (status) {

	    case 1:
                printf("Lower search bound (%.g) exceeded.\n", bound);
		break;

	    case 2:
                printf("Upper search bound (%.g) exceeded.\n", bound);
		break;

	    default:
		if (status < 0) {
                    static char *arghhh[] = { "??", "??WHICH??", "-P/-Q",
                                               "-Q/-P", "-X", "-D" };

		    if (status < -5) {
			status = -5;
		    }
                    printf("Argument %s out of range.\n", arghhh[-status]);
		} else {
                    printf("Error in calculation, status %d.\n", status);
		}
		break;
	}
	return 1;
    }

    return 0;
}
