/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.awt.*;

public class Bigraph extends VisualFeedback {
    private int v_ntrials = 0, v_left, v_right;
    private Color c_chart, c_pos, c_neg;
    private Image chart;
    
    // Constructors

    public Bigraph(int width, int height) {
        super(width, height);
        c_chart = null;
        c_pos = Color.blue;
        c_neg = Color.red;
    }

    /*  SETVIEW  --  Set view of binomial distribution  */

    public void setView(int ntrials,  // Number of trials in run
                        int left,     // Number of hits at left of window
                        int right)    // Number of hits at right of window
    {
        if (chart != null) {
            chart.flush();
            chart = null;
        }
        v_ntrials = ntrials;
        v_left = left;
        v_right = right;
    }

    /*  NORMALDISTRIBUTION  --  Calculate the probability density function
                                for the normal distribution with mean mu
                                and standard deviation sigma at a value x.
                                We use the normal distribution to plot  
                                the chart rather than the actual binomial
                                distribution to avoid "chunkiness" when
                                zoomed in on a portion of the curve.  */

    private static final double normalDistribution(double x, double mu, double sigma)
    {
        return (1 / Math.sqrt(2 * Math.PI * sigma)) *
                Math.exp(-((x - mu) * (x - mu)) / (2 * sigma * sigma));
    }

    //  PAINTWINDOW  --  Paint the component window
    
    public void paintWindow(Graphics g) {
        if (v_ntrials > 0) {
            int s;

            s = (((currentScore + (v_ntrials / 2)) - v_left) * size.width) / (v_right - v_left);

            /*  If we're about to run off the edge, pan in the appropriate
                direction.  */

            if (s < 0 || s >= size.width) {
                if (s < size.width / 2) {
                    setView(v_ntrials, v_left - size.width / 8,
                                       v_right - size.width / 8);
                } else {
                    setView(v_ntrials, v_left + size.width / 8,
                                       v_right + size.width / 8);
                }
                s = (((currentScore + (v_ntrials / 2)) - v_left) * size.width) / (v_right - v_left);
            }

            if (chart == null) {
                Graphics gfx;
                int x, y;
                double plot, stdev = Math.sqrt(v_ntrials / 4);
                double mscale = normalDistribution(v_ntrials / 2,
                    v_ntrials / 2, stdev);

                //  Create chart image

                chart = createImage(size.width, size.height);
                gfx = chart.getGraphics();
                gfx.setColor(c_background);
                gfx.fillRect(0, 0, size.width, size.height);

                if (c_chart != null) {
                    gfx.setColor(c_chart);
                }
                for (x = 0; x < size.width; x++) {
                    double sx;

                    plot = normalDistribution(sx = v_left + ((v_right - v_left) * x) / ((double) size.width),
                        v_ntrials / 2, stdev);

                    /*  If no colour is specified for the bell curve
                        chart, generate a "rainbow" plot in which
                        white fades to red and green as the significance
                        increases.  */
 
                    if (c_chart == null) {
                        double t;
                        int cr, cg, cb;

                        t = plot / mscale;

                        if (sx < (v_ntrials / 2)) {
                            cg = cb = (int) (t * 255);
                            cr = 255;
                        } else {
                            cr = cb = (int) (t * 255);
                            cg = 255;
                        }
                        gfx.setColor(new Color(cr, cg, cb));
                    }
                    y = (int) ((size.height - 1) - (size.height - 1) * plot / mscale);
                    gfx.drawLine(x, size.height - 1, x, y);
                }
            }

            g.drawImage(chart, 0, 0, this);

            //  Paint current score on chart

            if (s >= 0 && s < size.width) {
                g.setColor(currentScore >= 0 ? c_pos : c_neg);
                g.drawLine(s, 0, s, size.height - 1);
            }
        }
    }

    /*  VALIDATE  --  We need to know when the layout manager
                      changes our size so we can regenerate the
                      face bitmap on the next paint request.  */

    public void validate() {
        int dim;

        super.validate();
        size = new Dimension(size());
        chart = null;
    }
}
