#! /bin/perl

    #   PsiClock Request Processor
    #
    #   This Perl script is invoked by the HotBits CGI shell script
    #   as follows:
    #
    #       /bin/perl Psiclock.pl <mode> <E-mail> <goal> <sound> \
    #                             <experiment> <nbytes>
    #
    #   where the arguments are:
    #
    #       mode        Experiment mode: "Demo", "Practice", "Record"
    #       E-mail      User's E-mail address (for Record mode)
    #       goal        Goal: "Wide" or "Narrow" swing
    #       sound       "-s" if sound desired
    #       experiment  Name of Java feedback applet
    #       nbytes      Number of random bytes requested
    #
    #   For debugging, you can run the program directly from the
    #   command line and examine the HTML it returns to standard
    #   output.
    #
    #   This program is designed to be executed by Perl 4.036.

    #   Configuration

    #   Set $TestRand to 1 when debugging HTML output, log file
    #   generation, etc., to avoid depleting the HotBits inventory.
    #   It generates its own random data locally rather than
    #   contacting the server.

    $TestRand = 0;

    #   Any number of HotBits compatible sites can be interrogated in
    #   an attempt to satisfy the user's request.  The following
    #   arrays give the host names and ports on the servers to be
    #   interrogated.  The settings below are configured for running
    #   on www.fourmilab.ch, which we address as 'localhost' to avoid
    #   a potentially costly domain name lookup (due to in.named bloat
    #   due to Web traffic).  To run this program on another site but
    #   continue to obtain random data from the Fourmilab HotBits
    #   server, replace 'localhost' with 'www.fourmilab.ch'.  The port
    #   number remains the same.

    @hosts = ('localhost');
    @ports = (80);

    #   Log directory for Practice and Record runs.  The log file
    #   in this directory is written into <experiment>.dat.  Each
    #   log item contains the experiment name to facilitate analysis
    #   of results from one or more feedback programs.

    $logdir = "/files/server/logs/rpkp";
    $logfmt = '1';                    # Log format

    #   These are the correct values for Solaris 2.5; you may
    #   need to change them for other flavours of Unix.  You
    #   will usually find the definitions of these tags in
    #   /usr/include/sys/socket.h.

    $AF_INET = 2;
    $SOCK_STREAM = 2;

    require "shlock.pl";

    #   Generate error report if something goes wrong with the server

    sub errorOut {
    print <<"EOH"
Content-Type: text/html

<html>
<head>
<title>PsiClock Error</title>
</head>
<body>
<h1>Could not obtain random data</h1>
<hr>
<p>
The experiment you requested requires random data from the
<a href="http://www.fourmilab.ch/">Fourmilab</a>
<a href="http://www.fourmilab.ch/hotbits/">HotBits server</a>.
Unfortunately, the following error occurred while accessing
the server to obtain the random data:
<p>
<blockquote>
<b>$_[0]</b>
</blockquote>
<p>
The HotBits server may be temporarily down; please try again in a few
hours.  In the meanwhile, you can run experiments in Demo mode without
requiring data from the HotBits server.
<p>
<h2><a href=\"http://www.fourmilab.ch/rpkp/experiments/$Experiment/\">Back to experiment request</a></h2>
</body>
</html>
EOH
    ;
        exit(0);
    }

    #   Open temporary file

    sub main'open_temp {
        local($FH_name, $filename) = @_;
        local($inode1, $inode2) = ();

        # force unqualified filehandles into callers' package
        local($package) = caller;
        $FH_name =~ s/^[^']+$/$package'$&/;

        if ( -e $filename ) {
            &main'abort("Failed to open temp file $filename, it exists");
        }

        open($FH_name, ">> $filename") || 
            &main'abort("open of temp file $filename failed\n $!");

        if ( -l $filename ) {
            &main'abort("Temp file $filename is a symbolic link after opening");
        }

        if ( (stat(_))[3] != 1 ) {
            &main'abort("$filename has more than one link after opening");
        }

        $inode1 = (lstat(_))[1];
        local(*FH) = $FH_name;
        $inode2 = (stat(FH))[1];

        if ($inode1 != $inode2) {
            &main'abort("Inode for filename does not match filehandle!
                        Inode1=$inode1 Inode2=$inode2, Aborting.\n");
        }

        if ( (stat(FH))[3] != 1 ) {
            &main'abort("filehandle has more than one link after opening");
        }
        1;
    }

    #   Quote a string, inserting "\" escapes for single
    #   and quotes and "\"

    sub quote_string {
        local($s) = @_;
        $s =~ s/\\/\\\\/;
        $s =~ s/\"/\\\"/;
        $s =~ s/\'/\\\'/;
        '"' . $s . '"';
    }

    #   Process command line arguments

    ($Mode, $Email, $Goal, $Sound, $Experiment, $Nbytes) = @ARGV;

    #   Set defaults if arguments were not specified.  Since
    #   the shell script which invokes us always supplied the
    #   arguments, this is purely a convenience for debugging
    #   when the Perl program is invoked from the command line.

    if (!(defined $Mode)) {
        $Mode = 'Demo';
    }
    if (!(defined $Email)) {
        $Email = '';
    }
    if (!(defined $Goal)) {
        $Goal = '1';
    }
    if (!(defined $Sound)) {
        $Sound = '-n';
    }
    if (!(defined $Experiment)) {
        $Experiment = 'pendulum';
    }
    if (!(defined $Nbytes)) {
        $Nbytes = 128;
    }

    #   If a Record run is requested, make sure the user supplied
    #   an E-mail address or handle under which to record the results.

    if ($Mode eq 'Record' && $Email eq '') {
        print <<"EOH"
Content-Type: text/html

<html>
<head>
<title>Cannot record results</title>
</head>
<body>
<h1>Can't Record Results without User ID</h1>
<hr>
<p>
You have requested a "Record" mode run, the results of which will be
recorded under your name for statistical analysis, but you failed
to supply a your E-mail address (or a made-up "handle", if you
don't want to disclose your E-mail address).  Please return to the
experiment request page and enter your user identification when
performing a Record run.  You can perform Demo and Practice runs
(which are not included in the analysis of results) without entering
user identification.
<p>
<h2><a href=\"http://www.fourmilab.ch/rpkp/experiments/$Experiment/\">Back to experiment request</a></h2>
</body>
</html>
EOH
        ;
        exit(0);
    }


    #   If this is not a Demo run, contact a HotBits server to obtain
    #   the random data to be used by this run.

    if ($Mode ne 'Demo') {
        $ok = 0;

        if ($TestRand) {
            srand(time());
            $j = 0;
            for ($i = 0; $i < $Nbytes; $i++) {
                if ($j >= 32) {
                    $hexdata .= "\n";
                    $j = 0;
                }
                $hexdata .= sprintf("%02X", int(rand(256)));
                $j++;
            }
            $hexdata .= "\n";
            $ok = 1;
        } else {
            $sockaddr = 'S n a4 x8';

            #   All ready to go--now loop through the hosts and try
            #   to find one willing to fulfill our request.

            while (@hosts) {
                $them = shift(@hosts);
                $port = shift(@ports);
# print("Trying $them:$port\n");      # Uncomment to show polls of servers
                ($name, $aliases, $proto) = getprotobyname('tcp');
                ($name, $aliases, $type, $len, $thataddr) = gethostbyname($them);

                $that = pack($sockaddr, $AF_INET, $port, $thataddr);

                if (!socket(S, $AF_INET, $SOCK_STREAM, $proto)) {
                    next;
                }

                if (!connect(S, $that)) {
                    next;
                }

                select(S); $| = 1; select(STDOUT);

                print S("GET /cgi-bin/uncgi/Hotbits?nbytes=$Nbytes&fmt=hex HTTP/1.0\r\n\r\n");

                $indata = 0;
                while (<S>) {
# print;    # Uncomment to show reply from HotBits server
                    if ($indata) {
                        if (substr($_, 0, 1) eq '<') {
                            $ok = 1;
                            last;
                        }
                        $hexdata .= $_;
                    } else {
                        if (substr($_, 0, 5) eq "<pre>") {
                            $indata = 1;
                        }
                    }
                }
                close(S);
                if ($ok) {
                    last;
                }
            }
        }
        if (!$ok) {
            &errorOut($!);
            exit(0);
        }
    }

    #   If this is a non-Demo run, save the run information in the
    #   log file.  For statistical analysis, we save both Practice
    #   and Record runs, but do not consider Practice runs when
    #   evaluating a subject's performance.  Note that we record only
    #   the raw random data delivered to the feedback program and
    #   defer any scoring or analysis to programs which process the
    #   log file later on.  This preserves all the raw data and allows
    #   analysis in various ways, independent of the scoring assumptions
    #   used in the feedback program.

    #   The log file is written in CSV format, with the following
    #   fields:

    #       1:  Log format
    #       2:  1 = Live data, 0 = Internal debug (ignore)
    #       3:  Experiment name
    #       4:  Mode (Practice/Record)
    #       5:  Goal
    #       6:  Sound enabled?
    #       7:  Subject's E-mail address
    #       8:  Subject's IP address
    #       9:  Subject's host name, if known
    #      10:  Time and date of run (Unix time() format)
    #      11:  Random data for run

    if ($Mode ne 'Demo') {
        $locked = 0;
        for ($l = 0; $l < 30; $l++) {
            if (&shlock("/tmp/RPKP.LOCK")) {
                $locked = 1;
                last;
            }
            sleep(1);
        }
        if ($locked) {
            open(LF, ">>$logdir/rpkp.log");
            $bogus = $TestRand ? '0' : '1';
            $td = time();
            $em = &quote_string($Email);
            print(LF "$logfmt,$bogus,$Experiment,$Mode,$Goal,$Sound,$em,");
            $em = &quote_string($ENV{'REMOTE_HOST'});
            printf(LF "$ENV{'REMOTE_ADDR'},$em,$td,");
            $s = $hexdata;
            $s =~ s/\n//g;
            print(LF $s);
            print(LF "\n");
            close(LF);
            unlink("/tmp/RPKP.LOCK");
        } else {
            if ($Mode eq 'Record') {
                print <<"EOH"
Content-Type: text/html

<html>
<head>
<title>PsiClock Internal Error</title>
</head>
<body>
<h1>Could not lock results file</h1>
<hr>
<p>
You have requested a "Record" mode run, the results of which will be
recorded under your name for statistical analysis.  Unfortunately,
the RPKP experiment was unable to obtain exclusive access to the
master log file in order to record your results.  This may be a
transient overload condition which won't recur if you request the
experiment again, or it may be due to a problem with the RPKP server,
in which case it probably will not be remedied for several hours.
<p>
Please try resubmitting your request.  If it fails a second time, please
try again tomorrow.
<p>
<h2><a href=\"http://www.fourmilab.ch/rpkp/experiments/$Experiment/\">Back to experiment request</a></h2>
</body>
</html>
EOH
            ;
            exit(0);
        }
    }
    }

    #   Now that we have the random data (if we, in fact, need
    #   it), emit the document for the experiment, passing the
    #   random data to the Java feedback applet.

    print "Content-type: text/html\r\n\r\n";

    print <<"EOH"
<html>
<head>
<title>RPKP $Mode Experiment</title>
</head>
<body>
<h1>RPKP $Mode Experiment</h1>
<hr>
<p>
EOH
    ;

    if (0) {
        print <<"EOH"
Arguments:
<p>
<pre>
    Mode:       |$Mode|
    Email:      |$Email|
    Goal:       |$Goal|
    Sound:      |$Sound|
    Experiment: |$Experiment|
    Nbytes:     |$Nbytes|

    REMOTE_HOST:    $ENV{'REMOTE_HOST'}
    REMOTE_ADDR:    $ENV{'REMOTE_ADDR'}
    REMOTE_USER:    $ENV{'REMOTE_USER'}
    REMOTE_IDENT:   $ENV{'REMOTE_IDENT'}
</pre>
EOH
        ;
    }

    if ($TestRand && !($Mode eq 'Demo')) {
        print <<"EOH"
<blockquote>
<b>
<font size="+1" color="#FF0000">
WARNING!  The CGI server program which produced this experiment was
compiled with the test variable "\$TestRand" nonzero.  The data for
this experiment were, therefore, generated programmatically rather
than by the hardware random number generator.  Results from this experiment
will <em>not</em> be included in the statistical analysis of results.
</font>
</b>
</blockquote>
EOH
        ;
    }

#if (0) {
#   Debug--dump whole environment
#print("<h3>Environment</h3>\n<pre>\n");
#print(`printenv`);
#print("</pre>\n");
#}

    $wgoal = $Goal;
    if (substr($wgoal, 1, 1) eq ':') {
        $wgoal = substr($wgoal, 2);
    }
    print <<"EOH"
<center><h2>Goal: $wgoal</h2></center>
EOH
    ;

    if ($Experiment eq "bellcurve") {
        $Appsize = "width=480 height=320";
    } else {
        $Appsize = "width=300 height=400";
    }

    print <<"EOH"
<center>
<applet codebase="http://www.fourmilab.ch/rpkp/experiments/$Experiment" code="$Experiment.class" $Appsize>
EOH
    ;

    $snd = "";
    if ($Sound eq '-s') {
        $snd = "\n<param name=sound value=\"-s\">";
    }

    print <<"EOH"
<param name=mode value="$Mode">
<param name=goal value="$Goal">$snd
<param name=version value="on">
<param name=score value="on">
<param name=nrand value="$Nbytes">
EOH
    ;

    #   Gnarly per-experiment parameter settings

    if ($Experiment eq "pendulum") {
        print <<"EOH"
<param name=background value="000000">
<param name=string value="808080">
<param name=bob value="ffff00">
EOH
        ;
    } elsif ($Experiment eq "horloge") {
        print <<"EOH"
<param name=background value="C0C0C0">
EOH
        ;
    }

    if (defined $hexdata) {
        print("<param name=randomdata value=\"\n$hexdata\">\n");
    }
    print <<"EOH"
<b>Sorry--it appears your browser doesn't understand Java applets.</b>
</applet>
</center>
EOH
    ;

    if ($Email ne '' && $Mode eq 'Record') {
        print <<"EOH"
<h3><a href=\"/cgi-bin/uncgi/RPKPreport?html=-h&all=&email=kelvin@fourmilab.ch\">Show My Experiment Log</a></h3>
EOH
        ;
    }

    print <<"EOH"

<h3><a href=\"http://www.fourmilab.ch/rpkp/experiments/$Experiment/\">Run this experiment again</a></h3>
<h3><a href=\"http://www.fourmilab.ch/rpkp/experiments/\">Other experiments</a></h3>

</body>
</html>
EOH
    ;
