/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.applet.*;
import java.awt.*;
import java.util.*;
import Scoring;

public class VisualFeedback extends Canvas {
    protected int cWidth, cHeight;
    protected AudioClip sound = null;
    protected int currentScore = 0;
    protected Color c_background;
    protected Dimension size;
    protected Scoring finalScore = null;
    protected int maxamp = Integer.MIN_VALUE, minamp = Integer.MAX_VALUE;
    protected boolean userQuit = false;
    static public String[][] parameters = {
        { "Background", "RRGGBB", "Background colour, RGB hex, default this.GetBackground()" },
    };
    
    // Constructors

    public VisualFeedback(int width, int height) {
        cWidth = width;
        cHeight = height;
        c_background = Color.lightGray;
        setSize(width, height);
    }

    //  Parameter setting methods

    /*  PARAMETERCOUNT  --  Return number of visual feedback
                            parameters.  */

    public int parameterCount() {
        return parameters.length;
    }

    /*  PARAMETERNAME  --  Return visual feedback parameter name, or
                           null if index outside parameter list.  */

    public String parameterName(int idx) {
        return (idx < parameters.length) ? parameters[idx][0] : null;
    }

    /*  PARSECOLOUR  --  Parse a hexadecimal string as a colour
                         specification, returning the default if
                         the parse fails.  This is provided as a
                         service to derived classes which need to
                         handle colour specifications.  */

    protected static Color parseColour(String s, Color c_def) {
        Integer i;
        Color c;

        if (s != null) {
            try {
                i = Integer.valueOf(s, 16);
                c = new Color(i.intValue());
            } catch (NumberFormatException e) {
                c = c_def;
            }
        } else {
            c = c_def;
        }
        return c;
    }

    /*  SETPARAMETER  --  Set a parameter with a value supplied by
                          the applet that created us.  If the
                          parameter was not passed to the applet,
                          pvalue is null.  Returns true if the
                          parameter index is known, false otherwise.  */

    public boolean setParameter(int pindex, String pvalue)
    {
        switch (pindex) {
            case 0:
                c_background = parseColour(pvalue, getBackground());
                return true;
        }
        return false;
    }

    /*  SETAUDIOCLIP  --  Specify the audio clip to be played (if any)
                          when the limit is reached.  */

    public void setAudioClip(AudioClip clip) {
        sound = clip;
    }

    /*  AUDIOUPDATE  --  Given current score and historical minimum
                         and maximum amplitudes, decide whether to
                         play the audio feedback clip.  Made
                         distinct from setRunningScore so subclasses
                         can change audio feedback strategy without
                         replicating logic in that method.  */

    protected void audioUpdate() {
        if (sound != null) {
            if (currentScore < minamp) {
                minamp = currentScore;
                sound.play();
            } else if (currentScore > maxamp) {
                maxamp = currentScore;
                sound.play();
            }
        }
    }

    /*  SETRUNNINGSCORE  --  Passed the score so far (zero == chance
                             expectation.  We don't perform the
                             optimisation of only updating the display
                             when the score changes, since that might
                             interfere with smooth animation on the
                             part of the display component.  */

    public void setRunningScore(int score) {
        currentScore = score;
        change();
        audioUpdate();
    }

    /*  POSTSCORE  --  Mark run complete and show results
                       scoreboard.  If quit is set, the user
                       aborted the run with the quit button.  */

    public void postScore(Scoring s, boolean quit)
    {
        finalScore = s;
        userQuit = quit;
        change();
    }

    /*  SETSIZE  --  Change the size of the display component.  */

    public void setSize(int width, int height) {
        cWidth = width;
        cHeight = height;
        size = new Dimension(width, height);
    }

    //  Layout manager interface functions

    //  PREFERREDSIZE  --  Indicate how large we'd like to be

    public Dimension preferredSize() {
        return new Dimension(cWidth, cHeight);
    }
    
    //  MINIMUMSIZE  --  Indicate the minimum we can squeeze into

    public Dimension minimumSize() {
        return new Dimension(160, 160);
    }

    //  CHANGE --  Update the component's state

    protected void change() {
        repaint();
    }

    /*  UPDATE  --  Update the image.  We override the default
                    implementation since we know we're repainting
                    the entire component every time and thus don't
                    need the clear to background colour before the
                    paint.  Using the default would result in flicker
                    during image update.  */

    public void update(Graphics g) {
        paint(g);
    }

    /*  PAINTWINDOW  --  Generate the visual feedback image.
                         This is separate from paint() to allow it
                         to be overridden in derived classes without
                         replicating the final status display logic
                         in paint().  */

    protected void paintWindow(Graphics g) {
        String s = "Score: " + currentScore;
        FontMetrics fm = g.getFontMetrics();
        int l;

        l = (cHeight - fm.getHeight()) / 2;

        g.setColor(c_background);
        g.fillRect(0, 0, cWidth, cHeight);
        g.setColor(currentScore == 0 ? Color.black :
            (currentScore > 0 ? Color.blue : Color.red));
        g.drawString(s, (cWidth - fm.stringWidth(s)) / 2, l);
    }

    //  PAINT  --  Paint the component
    
    public void paint(Graphics g) {

        if (userQuit) {
            finalScore.paintQuit(g, size.width, size.height);
            return;
        } else if (finalScore != null) {
            finalScore.paintScore(g, size.width, size.height);
            return;
        }
        paintWindow(g);
    }
}
