/*

                         Bar Meter Component

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.applet.*;
import java.awt.*;

class barMeterItem {
    public Color colour;              // Paint colour
    public int from, to;              // Extents in frame, 0 - FULLSCALE
    public int span;                  // Span within meter, 0 - FULLSCALE
    public int threeD;                // 3D draw: 0 = 2D, 1 = raise, 2 = lower
    boolean visible;                  // Is bar visible ?
};

public class barMeter extends Canvas {

    //  Constants used as arguments specifications

    public final static boolean VERTICAL = true,
                                HORIZONTAL = false;
    public final static int FLAT = 0, RAISED = 1, LOWERED = 2;
    public final static boolean VISIBLE = true, HIDDEN = false;
    public final static int FULLSCALE = 100000;

    private int cWidth, cHeight,      // Active region size
                fWidth = -1, fHeight, // Full width including border
                rWidth, rHeight,      // Requested width and height
                xBorder = 0, yBorder = 0; // Border size
    private barMeterItem bi[];        // Bar items in paint order
    private boolean isVertical;       // Is bar vertically oriented ?
    private boolean dirtyTrick;       // Hidden argument for paint speedup
    
    // Constructors

    public barMeter(int width, int height, boolean vertical) {
        setSize(width, height);
        isVertical = vertical;
        dirtyTrick = false;
    }

    /*  SETSIZE  --  Change the size of the display component.  */

    public void setSize(int width, int height) {
        cWidth = rWidth = width;
        cHeight = rHeight = height;
        dirtyTrick = false;
    }

    //  SETBORDER  --  Set borders around control

    public void setBorder(int xborder, int yborder) {
        xBorder = xborder;
        yBorder = yborder;
        dirtyTrick = false;
    }

    /*  ADDBAR  --  Add a bar to the bar meter.  Returns integer
                    handle for the bar.  */

    public int addBar(Color colour, int from, int to,
                      int span, int threeD, boolean visible) {
        barMeterItem b = new barMeterItem();
        barMeterItem bn[];
        int index;

        b.colour = colour;
        b.from = from;
        b.to = to;
        b.span = span;
        b.threeD = threeD;
        b.visible = visible;
        if (bi == null) {
            bi = new barMeterItem[0];
        }
        index = bi.length;
        bn = new barMeterItem[bi.length + 1];
        System.arraycopy(bi, 0, bn, 0, bi.length);
        bn[index] = b;
        bi = bn;
        dirtyTrick = false;
        return index;
    }

    //  Utility methods

    //  PERCENT  --  Return percentage of full-scale value

    static public int percent(int p) {       // Percent of number 0 to 100
        return (p * FULLSCALE) / 100;
    }

    static public int percent(int value, int scale) { // Percent of value along scale
        return (value * FULLSCALE) / scale;
    }

    static public int percent(double p) {    // Percent of double between 0 and 1
        return (int) (p * FULLSCALE);
    }

    /*  PIXELS  --  Returns scaled percentage corresponding to
                    a given number of pixels.  This is handy when
                    setting up borders and arranging bars within
                    one another.  */

    public int pixelsX(int p) {       // Pixels along X direction
        return (p * FULLSCALE) / cWidth;
    }
    public int pixelsY(int p) {       // Pixels along Y direction
        return (p * FULLSCALE) / cHeight;
    }

    /*  Update methods.  All repaint only if a visible change
        has occurred.  */

    //  MOVEBAR  --  Change the extents of a bar.

    public void moveBar(int handleBar, int from, int to) {
        if (bi[handleBar].from != from ||
            bi[handleBar].to != to) {

            /* Since a common application of this component is
               a progress meter, where a single bar advances as
               an operation is performed, we check for the special
               case in which ther bar being updated is the last in
               the draw order and its new extent is larger than the
               previous.  If this is the case, we set the flag
               dirtyTrick before forcing the repaint of the control,
               which skips painting any but the last control,
               thereby speeding up the paint but, more importantly,
               avoiding momentary screen flicker. */

            if ((handleBar == (bi.length - 1)) &&
                (bi[handleBar].from >= from) &&
                (bi[handleBar].to <= to)) {
                dirtyTrick = true;
            }
            bi[handleBar].from = from;
            bi[handleBar].to = to;
            if (bi[handleBar].visible) {
                if (dirtyTrick) {
                    paint(getGraphics());
                    dirtyTrick = false;
                } else {
                    repaint();
                }
            }
        }
    }

    //  SHOWBAR  --  Set bar visibility status

    public void showBar(int handleBar, boolean visible) {
        if (bi[handleBar].visible != visible) {
            bi[handleBar].visible = visible;
            repaint();
        }
    }

    //  Layout manager interface functions

    //  PREFERREDSIZE  --  Indicate how large we'd like to be

    public Dimension preferredSize() {
        return new Dimension(rWidth, rHeight);
    }
    
    //  MINIMUMSIZE  --  Indicate the minimum we can squeeze into

    public Dimension minimumSize() {
        return preferredSize();
    }

    //  CHANGE --  Update the component's state

    public void change() {
        repaint();
    }

    //  UPDATE  --  Update the image

    public void update(Graphics g) {
        paint(g);
    }

    //  TOSTRING  --  Show bar meter information

    public String toString() {
        String s = "barMeter with ";

        if (bi == null) {
            s += "no";
        } else {
            s += bi.length;
        }
        s += isVertical ? " vertical" : " horizontal";
        s += " bars";
        return s;
    }

    //  PAINT  --  Paint the component
    
    public void paint(Graphics p) {
        Rectangle r = bounds();

        if (r.width != fWidth || r.height != fHeight) {
            fWidth = r.width;
            fHeight = r.height;
            cWidth = fWidth - (xBorder * 2);
            cHeight = fHeight - (yBorder * 2);
            dirtyTrick = false;
        }
        if (bi != null) {
            int i;
            int extent = isVertical ? cHeight : cWidth,
                span = isVertical ? cWidth : cHeight;

            for (i = dirtyTrick ? (bi.length - 1) : 0; i < bi.length; i++) {
                if (bi[i].visible) {
                    int f = (extent * bi[i].from) / FULLSCALE,
                        t = (extent * bi[i].to) / FULLSCALE,
                        sw = (span * bi[i].span) / FULLSCALE,
                        so = (span - sw) / 2;
                    p.setColor(bi[i].colour);
                    if (isVertical) {
                        if (bi[i].threeD > 0) {
                            p.fill3DRect(xBorder + so, yBorder + f, sw, t - f,
                                         bi[i].threeD == 1);
                        } else {
                            p.fillRect(xBorder + so, yBorder + f, sw, t - f);
                        }
                    } else {
                        if (bi[i].threeD > 0) {
                            p.fill3DRect(xBorder + f, yBorder + so, t - f, sw,
                                         bi[i].threeD == 1);
                        } else {
                             p.fillRect(xBorder + f, yBorder + so, t - f, sw);
                        }
                    }
                }
            }
        }
    }
}
