/*
                              runBitSet

             Extension of BitSet for experiment data sets

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

    This module contains both the mechanics for manipulating bit
    sets and analysis functions which interpret the bit set as a
    series of independent tests each with a probability of success
    of 0.5, with statistics, therefore, of a binomial distribution.

    The BinomialDistribution package is used to calculate
    probabilities.  Lines prefixed with "//--" are definitions of
    statistical quantities usable with *any* distribution.  They are
    followed by the much shorter and faster special case for
    a binomial distribution.

*/

import java.util.*;
import BinomialDistribution;

public class runBitSet
{
    BitSet b;

    /*  Create runBitSet from hexadecimal run data.  The
        run data may contain white space, which is ignored.  */

    public runBitSet(String h)
    {
        StringBuffer sh = new StringBuffer(h.length());
        int i, n = 0;
        String s;

        //  Squash white space out of string
        for (i = 0; i < h.length(); i++) {
            if (!Character.isSpace(h.charAt(i))) {
                sh.append(h.charAt(i));
            }
        }

        //  Set bits from hex digit pairs
        s = sh.toString();
        b = new BitSet(4 * s.length());
        for (i = 0; i < s.length(); i++) {
            String c = s.substring(i, i + 1);
            int v = Integer.parseInt(c, 16);
            int m;

            for (m = 8; m > 0; m >>= 1) {
                if ((v & m) != 0) {
                    b.set(n);
                }
                n++;
            }
        }
    }

    /*  Create pseudorandom runBitSet from Java's Random class,   
        either based on the time or from a specified seed.  */

    private final void pseudorandomBitSet(int nbits, Random r)
    {
        int i;

        b = new BitSet(nbits);

        for (i = 0; i < nbits; i++) {
            if ((r.nextInt() & 0x100) == 0) {
                b.set(i);
            }
        }
    }

    public runBitSet(int nbits)
    {
        pseudorandomBitSet(nbits, new Random());
    }

    public runBitSet(int nbits, long seed)
    {
        pseudorandomBitSet(nbits, new Random(seed));
    }

    //  Convert vector to hex string

    public String toString()
    {
        int sz = b.size();
        int i, j;
        StringBuffer sb = new StringBuffer();

        for (i = 0; i < sz ; i += 4) {
            int n = 0;

            for (j = 0; j < 4; j++) {
                n += b.get(i + j) ? (1 << (3 - j)) : 0;
            }
            sb.append(Integer.toString(n, 16).toUpperCase());
        }
        return sb.toString();
    }

    //  Return number of bits in vector

    public int size()
    {
        return b.size();
    }

    //  Return bit from vector

    public boolean get(int n)
    {
        return b.get(n);
    }

    //  Count zero bits in vector

    public int numZero()
    {
        int i, n = 0;

        for (i = 0; i < b.size(); i++) {
            if (!b.get(i)) {
                n++;
            }
        }
        return n;
    }

    //  Count one bits in vector

    public int numOne()
    {
        return b.size() - numZero();
    }

    //  Binomial distribution calculations based on contents of set

    //  Mean (expected value)

    public int mean()
    {
        return size() / 2;
    }

    //  Probability mass for a given number of successes

    public double probability(int successes)
    {
        return BinomialDistribution.probability(successes, size(), 0.5);
    }

    //  Deviation of a given number of successes from the expected value

    public double deviation(int successes)
    {
        return Math.abs(mean() - successes);
    }

//--//  Variance term for a given number of successes
//--
//--public double variance(int successes)
//--{
//--    return probability(successes) *
//--           deviation(successes) * deviation(successes);
//--}

    //  Variance of distribution (sum of variance terms)

    public double variance()
    {
//--    int i;
//--    double v = 0;
//--
//--    for (i = 0; i < size(); i++) {
//--        v += variance(i);
//--    }
//--    return v;
/**/    return size() / 4;
    }

    //  Standard deviation of distribution (square root of variance)

    public double standardDeviation()
    {
        return Math.sqrt(variance());
    }

    /*  How many standard deviations does a number of successes
        diverge from the expected value?  */

    public double standardDeviation(int successes)
    {
        return deviation(successes) / standardDeviation();
    }

    /*  What is the probability of a given number of successes
        occurring by chance?  (Area under the probability distribution
        curve.)  This is done iteratively and is consequently very
        slow.  If you need to compute a lot of these, it would
        be better to use a closed-form approximation based on the
        limiting case of a normal distribution.  */

    public double significance(int successes)
    {
        int i;
        double s = 0;

        for (i = 0; i <= (mean() - deviation(successes)); i++) {
            s += probability(i);
        }
        return s;
    }

    /*  Calculate statistics of the actual bits in the set.
        We don't need to know what the goal was because the
        statistics are symmetrical around the mean.  */

    public double resultStandardDeviation()
    {
        return standardDeviation(numZero());
    }

    public double resultSignificance()
    {
        return significance(numZero());
    }

/*
    //  Test internal operations

    public void test()
    {
        int i;

        System.out.println("Hex: " + toString());
        System.out.println("Size: " + size() + ", Zeroes: " + numZero() + " Ones: " + numOne());
        System.out.println("Mean: " + mean());
        System.out.println("Variance: " + variance());
        System.out.println("Standard deviation: " + standardDeviation());
        System.out.println();
        System.out.println("Result: " + resultStandardDeviation() +
                           " sigma, probability: " + resultSignificance());

        System.out.println();
        for (i = 0; i < size(); i++) {
            System.out.println("Hits: " + i +
               "  Deviation: " + deviation(i) +
               "  Probability: " + probability(i) +
               "  Std dev: " + standardDeviation(i) +
               "  Significance: " + significance(i));
        }
    }

    public static void main(String args[]) {
        runBitSet b;

        b = new runBitSet("1F68ACB2 25250C4A    7357083BCC78996E19 6A5EB 3FCAB  5FD77F903B543BB  9B2BCDA899664DE55021D98E26FE153BF698AF34AE22D618B49D9271C338EF45B0BA5147B529218AF48645358EB608FC9DF720C9257A3AC55D5658EBE78217CD36CD0B66931630AFD1F0D3E6F08663D941723573E050E2060D2294EE4C2CDF6662AAB");

        b.test();
    }
/**/
}
