/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.awt.*;

public class Pendule extends VisualFeedback {

    //  Animation frames per swing of pendulum

    private static final int steps = 20;

    /*  Divergence from most probable value resulting in
        swing to edge of window or (negative) stationary
        pendulum.  */

    private static final int ampscore = 20;

    private double phase = 0;
    private int lx = -1, ly, lw = -1, ldx, ox, oy, length, bob, hub;
    private Color c_hub, c_string,    // Subcomponent colours
                  c_sdark, c_bob;
    private Image work = null;        // Offscreen work image
    private Graphics g;               // Graphics context for this image
    private Image i_bob;              // Image of pendulum bob
    private Image i_hub;              // Image of hub
    private int amplitude, ampmax;    // Swing amplitude and max amplitude
    private boolean armed = true;     // Is audio feedback armed ?

    // Constructor

    public Pendule(int width, int height) {
        super(width, height);
        c_hub = Color.green;
        c_string = Color.gray;
        c_sdark = c_string.darker();
        c_bob = Color.yellow;
    }

    //  CHANGE --  Update the component's state

    protected void change() {
        phase += Math.PI / steps;
        super.change();
    }

    /*  AUDIOUPDATE  --  Override default audio feedback logic, which
                         is performed in paintWindow in our case.  */

    protected final void audioUpdate() {
    }

    /*  SCORETOAMPLITUDE  --  Function which maps score, in terms of
                              excess of one or zero bits from the expected
                              value of zero, into amplitude of the pendulum's
                              swing.  */

    static final int scoreToAmplitude(int winWidth, int score) {
        return (winWidth / 4) + ((score * (winWidth / 4)) / ampscore);
    }

    //  BALLIMAGE  --  Create an image of a rounded ball

    protected Image ballImage(int radius, Color cball, Color cback) {
        Image i_ball;
        Graphics b;
        int cr = cball.getRed(), cg = cball.getGreen(),
            cb = cball.getBlue(), i;

        i_ball = createImage(radius * 2, radius * 2);
        b = i_ball.getGraphics();
        b.setColor(cback);
        b.fillRect(0, 0, radius * 2, radius * 2);
        for (i = radius; i > 0; i--) {
            double shade = Math.cos((Math.PI * 0.42) * ((double) i) / radius);

            b.setColor(new Color((int) (cr * shade), (int) (cg * shade),
                                 (int) (cb * shade)));
            b.fillOval(radius - i, radius - i, i * 2, i * 2);
        }
        b.dispose();
        return i_ball;
    }

    //  PAINTWINDOW  --  Paint the component
    
    public void paintWindow(Graphics p) {
        int dx, dy;
        Rectangle r = bounds();

        amplitude = scoreToAmplitude(r.width, currentScore);
        if (work == null) {
            lw = r.width;
            work = createImage(r.width, r.height);
            g = work.getGraphics();

            //  Auto-scale the display to the component size

            ox = r.width / 2;
            oy = r.height / 8;
            bob = Math.min(r.height, r.width) / 12;
            length = (r.height * 7) / 8;
            hub = bob / 4;
            ampmax = scoreToAmplitude(r.width, ampscore) - bob;

            //  Create the image for the bob

            i_bob = ballImage(bob, c_bob, c_background);

            //  Create the image for the hub

            i_hub = ballImage(hub, c_hub, c_background);
        }

        /*  Now draw the pendulum, using the hub and bob images
            already created, and drawing the string on the fly. */

        //  Draw in auxiliary bitmap and copy to screen.

        g.setColor(c_background);
        g.fillRect(0, 0, r.width, r.height);
        dx = (int) (amplitude * Math.sin(phase));
        dy = (int) Math.sqrt(length * length - dx * dx);
        g.setColor(c_string);
        g.drawLine(ox, oy, lx = ox + dx, ly = dy);
        g.setColor(c_sdark);
        g.drawLine(ox - 1, oy, lx - 1, ly);
        g.drawLine(ox + 1, oy, lx + 1, ly);
        g.drawImage(i_hub, ox - hub, oy - hub, this);
        g.drawImage(i_bob, lx - bob, ly - bob, this);
        p.drawImage(work, 0, 0, this);

        /*  Our audio feedback is different from the default behaviour
            of the superclass, which plays the feedback sound every time
            a new high or low score is made during the run.  We 
            override audioUpdate() with a method that does nothing
            and, instead, generate the feedback below.  The feeback
            sounds whenever a wide-swinging pendulum hits the left
            or right edge of the window.  Once this happens, the
            audio is disarmed until the pendulum passes through
            the vertical, to avoid extra bonks if there's jitter
            near the edge.  */

        if (armed) {
            if (sound != null && ((int) (Math.abs(dx) + 0.5) >= ampmax)) {
                sound.play();
                armed = false;
                ldx = dx;
            }
        } else {
            //  Re-arm sound at zero crossing
            if ((ldx * dx) < 0) {
                armed = true;
            }
        }
    }

    /*  VALIDATE  --  We need to know when the layout manager
                      changes our size so we can regenerate the
                      face bitmap on the next paint request.  */

    public void validate() {
        super.validate();
        size = new Dimension(size().width, size().height);
        work = null;
    }
}
