/*

               by John Walker  --  kelvin@fourmilab.ch
                       http://www.fourmilab.ch/

                This program is in the public domain.

*/

import java.awt.*;

public class Organ extends VisualFeedback {
    private boolean initialised = false, forcePaint = false,
                    newScore = false;
    private Color c_chart, c_pos, c_neg;
    private Image chart;

    int bitsPerExp = 32;
    private int totalExp = 0, lhmax = -1;
    private int[] histo, ohisto;
    
    // Constructors

    public Organ(int width, int height) {
        super(width, height);
        c_chart = null;
        c_pos = Color.blue;
        c_neg = Color.red;
        reset();
    }

    /*  SETVIEW  --  Set view of binomial distribution  */

    public void setView() {
        if (chart != null) {
            chart.flush();
            chart = null;
        }
        initialised = true;
    }

    //  RESET  --  Reset parameters for a new run

    public void reset() {
        if (chart != null) {
            chart.flush();
            chart = null;
        }
        histo = new int[bitsPerExp + 1];
        ohisto = new int[bitsPerExp + 1];
        totalExp = 0;
        lhmax = -1;
//System.out.println("reset");
    }

    //  RESET  --  Reset and specify new bits per experiment

    public void reset(int n) {
        bitsPerExp = n;
        reset();
    }

    /*  REPAIR  --  Repair damage when AWT doesn't let know we
                    been exposed or scrolled.  */

    public void repair() {
        forcePaint = true;
        repaint();
    }

    /*  NORMALDISTRIBUTION  --  Calculate the probability density function
                                for the normal distribution with mean mu
                                and standard deviation sigma at a value x.
                                We use the normal distribution to plot  
                                the chart rather than the actual binomial
                                distribution to avoid "chunkiness" when
                                zoomed in on a portion of the curve.  */

    private static final double normalDistribution(double x, double mu, double sigma)
    {
        return (1 / Math.sqrt(2 * Math.PI * sigma)) *
                Math.exp(-((x - mu) * (x - mu)) / (2 * sigma * sigma));
    }

    /*  SETRUNNINGSCORE  --  We need to override setRunningScore in
                             order to capture individual bits from
                             the generator (which we infer from the
                             changes in the score).  We can't do this
                             in paintWindow, since it may be called
                             for other reasons (resize, damage repair,
                             etc.) than the arrival of a new bit. */

    public void setRunningScore(int score) {
        newScore = true;
        histo[score]++;
        totalExp++;
        super.setRunningScore(score);
    }

    //  PAINTWINDOW  --  Paint the component window
    
    public void paintWindow(Graphics g) {
        if (initialised) {
            boolean update = !forcePaint && newScore;
            int i;
            Rectangle clip = g.getClipRect();

            forcePaint = false;
            newScore = false;

            /* This little test attempts to determine if we've been
               called to repair damage due to a window exposure or
               scrolling event.  Wouldn't it nice if it just *told*
               us when that happened? */

            if (clip.x != 0 || clip.y != 0 ||
                clip.width != size.width || clip.height != size.height) {
                update = false;
            }

            if (chart == null) {
                Graphics gfx;
                int x, y;
                double plot, stdev = Math.sqrt(bitsPerExp / 4);
                double mscale = normalDistribution(bitsPerExp / 2,
                    bitsPerExp / 2, stdev);

                //  Create chart image

                update = false;
                chart = createImage(size.width, size.height);
                gfx = chart.getGraphics();
                gfx.setColor(c_background);
                gfx.fillRect(0, 0, size.width, size.height);

                if (c_chart != null) {
                    gfx.setColor(c_chart);
                }
                for (x = 0; x < size.width; x++) {
                    double sx;

                    plot = normalDistribution(sx = (bitsPerExp * x) / ((double) size.width),
                        bitsPerExp / 2, stdev);

                    /*  If no colour is specified for the bell curve
                        chart, generate a "rainbow" plot in which
                        white fades to red and green as the significance
                        increases.  */
 
                    if (c_chart == null) {
                        double t;
                        int cr, cg, cb;

                        t = plot / mscale;

                        if (sx < (bitsPerExp / 2)) {
                            cg = cb = (int) (t * 255);
                            cr = 255;
                        } else {
                            cr = cb = (int) (t * 255);
                            cg = 255;
                        }
                        gfx.setColor(new Color(cr, cg, cb));
                    }
                    y = (int) ((size.height - 1) - (size.height - 1) * plot / mscale);
                    gfx.drawLine(x, size.height - 1, x, y);
                }
            }

            //  Paint histogram

            if (totalExp == 0) {
                g.drawImage(chart, 0, 0, this);
            } else {
                int barWid = Math.max(1, size.width / bitsPerExp);
                int hmax = -1;

                for (i = 0; i <= bitsPerExp; i++) {
                    hmax = Math.max(hmax, histo[i]);
                }
                if (lhmax != hmax) {
                    update = false;
                    lhmax = hmax;
                }

                if (!update) {
                    g.drawImage(chart, 0, 0, this);
                }
                g.setColor(c_pos);
                for (i = 0; i <= bitsPerExp; i++) {
                    if (histo[i] > 0 && (!update || (ohisto[i] != histo[i]))) {
                        int x = ((i * size.width) / bitsPerExp) - (barWid / 2),
                            y = ((size.height - 1) * histo[i]) / hmax;
                        g.fillRect(x + barWid / 4, size.height - y, barWid - (barWid / 2), y);
                    }
                }
                System.arraycopy(histo, 0, ohisto, 0, bitsPerExp);

                /* Paint registration mark at mean to check alignment
                {
                    int x = (((bitsPerExp / 2) * size.width) / bitsPerExp);

                    g.setColor(c_neg);
                    g.drawLine(x, 0, x, size.height - 1);
                }
                /**/
            }
        }
    }

    /*  VALIDATE  --  We need to know when the layout manager
                      changes our size so we can regenerate the
                      background bitmap on the next paint request.  */

    public void validate() {
        int dim;

        super.validate();
        size = new Dimension(size());
        if (chart != null) {
            chart.flush();
        }
        chart = null;
    }
}
